module SQLiterate
  module Node
    module DisjunctionExpression
      def tables
        e.tables + r.elements.flat_map { |e| e.e.tables }
      end
    end

    module ConjunctionExpression
      def tables
        e.tables + r.elements.flat_map { |e| e.e.tables }
      end
    end

    module NegationExpression
      def tables
        e.tables
      end
    end

    module EqualityExpression
      def tables
        e.tables + r.elements.flat_map { |e| e.e.tables }
      end
    end

    module BetweenExpression
      def tables
        e.tables + r.elements.flat_map do |e|
          e.l.tables + e.r.tables
        end
      end
    end

    module SetPredicateExpression
      def tables
        e.tables + set_constructor.tables
      end
    end

    module SetConstructor
      def tables
        e.tables
      end
    end

    module ExistsExpression
      def tables
        query_expression.tables
      end
    end

    module TestExpression
      module Binary
        def tables
          l.tables + r.tables
        end
      end
      module Unary
        def tables
          e.tables
        end
      end
    end

    module GenExpression
      def tables
        gen_value.tables + r.elements.flat_map { |e| e.gen_value.tables }
      end
    end

    module GenValue
      module Literal
        def tables
          []
        end
      end
      module Subscript
        def tables
          field_selection.tables + r.elements.flat_map do |e|
            e.range_expression.tables
          end
        end
      end
    end

    module CaseExpression
      def tables
        case_else_section.tables + \
        case_when_section.tables + r.elements.flat_map do |e|
          e.case_when_section.tables
        end
      end
    end

    module CaseWhenSection
      def tables
        w.tables + t.tables
      end
    end

    module CaseElseSection
      def tables
        respond_to?(:scalar_expression) ? scalar_expression.tables : []
      end
    end

    module FunctionCall
      def tables
        function_params.tables
      end
    end

    module TypeCast
      def tables
        scalar_expression.tables
      end
    end

    module VoidExpression
      def tables
        []
      end
    end

    module AggregateExpression
      def tables
        expressions_list.tables + order_by_clause.tables
      end
    end

    module OrderByClause
      def tables
        respond_to?(:sort_expressions_list) ? sort_expressions_list.tables : []
      end
    end

    module SortExpressionsList
      def tables
        sort_expression.tables + r.elements.flat_map do |e|
          e.sort_expression.tables
        end
      end
    end

    module SortExpression
      def tables
        scalar_expression.tables
      end
    end

    module ExpressionsList
      def tables
        scalar_expression.tables + r.elements.flat_map do |e|
          e.scalar_expression.tables
        end
      end
    end

    module NamedExpressionsList
      def tables
        named_expression.tables + r.elements.flat_map do |e|
          e.named_expression.tables
        end
      end
    end

    module NamedExpression
      def tables
        scalar_expression.tables
      end
    end

    module AllColumns
      def tables
        []
      end
    end

    module RangeExpression
      def tables
        b.tables + e.tables
      end
    end

    module FieldSelection
      def tables
        row_value.tables
      end
    end

    module RowValue
      module Query
        def tables
          query_expression.tables
        end
      end
      module Expression
        def tables
          scalar_expression.tables
        end
      end
      module Identifier
        def tables
          []
        end
      end
    end

    module QueryExpression
      def tables
        query_value.tables + r.elements.flat_map { |e| e.query_value.tables }
      end
    end

    module QueryValue
      module Expression
        def tables
          query_expression.tables
        end
      end
      module Select
      end
    end

    module SelectQuery
      def tables
        select_list.tables + \
        table_expression.tables + \
        order_by_clause.tables + \
        limit_clause.tables + \
        offset_clause.tables
      end
    end

    module TableExpression
      def tables
        from_clause.tables + where_clause.tables + group_by_clause.tables
      end
    end

    module FromClause
      def tables
        respond_to?(:table_references) ? table_references.tables : []
      end
    end

    module WhereClause
      def tables
        respond_to?(:scalar_expression) ? scalar_expression.tables : []
      end
    end

    module GroupByClause
      def tables
        if respond_to?(:expressions_list)
          expressions_list.tables + having_clause.tables
        else
          []
        end
      end
    end

    module HavingClause
      def tables
        respond_to?(:scalar_expression) ? scalar_expression.tables : []
      end
    end

    module TableReferences
      def tables
        table_joins.tables + r.elements.flat_map { |e| e.table_joins.tables }
      end
    end

    module TableJoins
      def tables
        [table_reference.table] + r.elements.map { |e| e.joined_table.table }
      end
    end

    module TableReference
      def table
        if respond_to?(:table_spec)
          [ table_name.name, table_spec.tables ]
        else
          [ name, [name] ]
        end
      end
    end

    module TableSpec
      module Query
        def tables
          query_expression.tables
        end
      end
      module Joins
        def tables
          table_joins.tables
        end
      end
      module Table
        def tables
          [name]
        end
      end
    end

    module Join
      def table
        table_reference.table
      end
    end

    module LimitClause
      def tables
        respond_to?(:scalar_expression) ? scalar_expression.tables : []
      end
    end

    module OffsetClause
      def tables
        respond_to?(:scalar_expression) ? scalar_expression.tables : []
      end
    end
  end
end
